@@ -40,12 +40,16 @@
 #if PG_VERSION_NUM >= 90600
 #include "postmaster/autovacuum.h"
 #endif
+#if PG_VERSION_NUM >= 120000
+#include "replication/walsender.h"
+#endif
 #include "storage/fd.h"
 #include "storage/ipc.h"
 #include "storage/spin.h"
 #include "utils/builtins.h"
 #include "utils/guc.h"
 
+#include "pg_stat_kcache.h"
 
 PG_MODULE_MAGIC;
 
@@ -63,21 +67,11 @@ typedef uint64 pgsk_queryid;
 typedef uint32 pgsk_queryid;
 #endif
 
-#define PG_STAT_KCACHE_COLS_V2_0	7
-#define PG_STAT_KCACHE_COLS_V2_1	15
-#define PG_STAT_KCACHE_COLS_V2_2	28
-#define PG_STAT_KCACHE_COLS			28	/* maximum of above */
-
 #define USAGE_INCREASE			(1.0)
 #define USAGE_DECREASE_FACTOR	(0.99)	/* decreased every pgsk_entry_dealloc */
 #define USAGE_DEALLOC_PERCENT	5		/* free this % of entries at once */
 #define USAGE_INIT				(1.0)	/* including initial planning */
 
-/* ru_inblock block size is 512 bytes with Linux
- * see http://lkml.indiana.edu/hypermail/linux/kernel/0703.2/0937.html
- */
-#define RUSAGE_BLOCK_SIZE	512			/* Size of a block for getrusage() */
-
 #define TIMEVAL_DIFF(start, end) ((double) end.tv_sec + (double) end.tv_usec / 1000000.0) \
 	- ((double) start.tv_sec + (double) start.tv_usec / 1000000.0)
 
@@ -97,19 +91,6 @@ typedef enum pgskVersion
 	PGSK_V2_2
 } pgskVersion;
 
-typedef enum pgskStoreKind
-{
-	/*
-	 * PGSS_PLAN and PGSS_EXEC must be respectively 0 and 1 as they're used to
-	 * reference the underlying values in the arrays in the Counters struct,
-	 * and this order is required in pg_stat_statements_internal().
-	 */
-	PGSK_PLAN = 0,
-	PGSK_EXEC,
-
-	PGSK_NUMKIND				/* Must be last value of this enum */
-} pgskStoreKind;
-
 static const uint32 PGSK_FILE_HEADER = 0x0d756e11;
 
 static struct	rusage exec_rusage_start[PGSK_MAX_NESTED_LEVEL];
@@ -117,39 +98,6 @@ static struct	rusage exec_rusage_start[PGSK_MAX_NESTED_LEVEL];
 static struct	rusage plan_rusage_start[PGSK_MAX_NESTED_LEVEL];
 #endif
 
-/*
- * Current getrusage counters.
- *
- * For platform without getrusage support, we rely on postgres implementation
- * defined in rusagestub.h, which only supports user and system time.
- *
- * Note that the following counters are not maintained on GNU/Linux:
- *   - ru_nswap
- *   - ru_msgsnd
- *   - ru_msgrcv
- *   - ru_nsignals
-*/
-typedef struct pgskCounters
-{
-	double			usage;		/* usage factor */
-	/* These fields are always used */
-	float8			utime;		/* CPU user time */
-	float8			stime;		/* CPU system time */
-#ifdef HAVE_GETRUSAGE
-	/* These fields are only used for platform with HAVE_GETRUSAGE defined */
-	int64			minflts;	/* page reclaims (soft page faults) */
-	int64			majflts;	/* page faults (hard page faults) */
-	int64			nswaps;		/* swaps */
-	int64			reads;		/* Physical block reads */
-	int64			writes;		/* Physical block writes */
-	int64			msgsnds;	/* IPC messages sent */
-	int64			msgrcvs;	/* IPC messages received */
-	int64			nsignals;	/* signals received */
-	int64			nvcsws;		/* voluntary context witches */
-	int64			nivcsws;	/* unvoluntary context witches */
-#endif
-} pgskCounters;
-
 static int	pgsk_max = 0;	/* max #queries to store. pg_stat_statements.max is used */
 
 /*
@@ -199,6 +147,9 @@ static int	plan_nested_level = 0;
 
 
 /* saved hook address in case of unload */
+#if PG_VERSION_NUM >= 150000
+static shmem_request_hook_type prev_shmem_request_hook = NULL;
+#endif
 static shmem_startup_hook_type prev_shmem_startup_hook = NULL;
 #if PG_VERSION_NUM >= 130000
 static planner_hook_type prev_planner_hook = NULL;
@@ -243,7 +194,6 @@ static bool pgsk_track_planning;	/* whether to track planning duration */
 /*--- Functions --- */
 
 void	_PG_init(void);
-void	_PG_fini(void);
 
 extern PGDLLEXPORT Datum	pg_stat_kcache_reset(PG_FUNCTION_ARGS);
 extern PGDLLEXPORT Datum	pg_stat_kcache(PG_FUNCTION_ARGS);
@@ -261,6 +211,9 @@ static void pg_stat_kcache_internal(FunctionCallInfo fcinfo, pgskVersion
 static void pgsk_setmax(void);
 static Size pgsk_memsize(void);
 
+#if PG_VERSION_NUM >= 150000
+static void pgsk_shmem_request(void);
+#endif
 static void pgsk_shmem_startup(void);
 static void pgsk_shmem_shutdown(int code, Datum arg);
 #if PG_VERSION_NUM >= 130000
@@ -357,14 +310,24 @@ _PG_init(void)
 
 	/* set pgsk_max if needed */
 	pgsk_setmax();
+	/*
+	 * If you change code here, don't forget to also report the modifications
+	 * in pgsk_shmem_request() for pg15 and later.
+	 */
+#if PG_VERSION_NUM < 150000
 	RequestAddinShmemSpace(pgsk_memsize());
 #if PG_VERSION_NUM >= 90600
 	RequestNamedLWLockTranche("pg_stat_kcache", 2);
 #else
 	RequestAddinLWLocks(1);
-#endif
+#endif		/* pg 9.6+ */
+#endif		/* pg 15- */
 
 	/* Install hook */
+#if PG_VERSION_NUM >= 150000
+	prev_shmem_request_hook = shmem_request_hook;
+	shmem_request_hook = pgsk_shmem_request;
+#endif
 	prev_shmem_startup_hook = shmem_startup_hook;
 	shmem_startup_hook = pgsk_shmem_startup;
 #if PG_VERSION_NUM >= 130000
@@ -381,20 +344,6 @@ _PG_init(void)
 	ExecutorEnd_hook = pgsk_ExecutorEnd;
 }
 
-void
-_PG_fini(void)
-{
-	/* uninstall hook */
-	shmem_startup_hook = prev_shmem_startup_hook;
-#if PG_VERSION_NUM >= 130000
-	planner_hook = prev_planner_hook;
-#endif
-	ExecutorStart_hook = prev_ExecutorStart;
-	ExecutorRun_hook = prev_ExecutorRun;
-	ExecutorFinish_hook = prev_ExecutorFinish;
-	ExecutorEnd_hook = prev_ExecutorEnd;
-}
-
 static bool
 pgsk_assign_linux_hz_check_hook(int *newval, void **extra, GucSource source)
 {
@@ -474,6 +423,24 @@ pgsk_set_queryid(pgsk_queryid queryid)
 }
 #endif
 
+#if PG_VERSION_NUM >= 150000
+/*
+ * Request additional shared memory resources.
+ *
+ * If you change code here, don't forget to also report the modifications in
+ * _PG_init() for pg14 and below.
+ */
+static void
+pgsk_shmem_request(void)
+{
+	if (prev_shmem_request_hook)
+		prev_shmem_request_hook();
+
+	RequestAddinShmemSpace(pgsk_memsize());
+	RequestNamedLWLockTranche("pg_stat_kcache", 2);
+}
+#endif
+
 static void
 pgsk_shmem_startup(void)
 {
@@ -720,12 +687,23 @@ pgsk_queryids_array_size(void)
 {
 	/*
 	 * queryid isn't pushed to parallel workers.  We store them in shared mem
-	 * for each query, identified by their BackendId.  If need room for all
-	 * possible backends, plus autovacuum launcher and workers, plus bg workers
-	 * and an extra one since BackendId numerotation starts at 1.
+	 * for each query, identified by their BackendId.  It therefore needs room
+	 * for all possible backends, plus autovacuum launcher and workers, plus bg
+	 * workers and an extra one since BackendId numerotation starts at 1.
+	 * Starting with pg12, wal senders aren't part of MaxConnections anymore,
+	 * so they need to be accounted for.
 	 */
+#if PG_VERSION_NUM >= 150000
+	Assert (MaxBackends > 0);
+	return (sizeof(pgsk_queryid) * (MaxBackends + 1));
+#else
 	return (sizeof(pgsk_queryid) * (MaxConnections + autovacuum_max_workers + 1
-							+ max_worker_processes + 1));
+							+ max_worker_processes
+#if PG_VERSION_NUM >= 120000
+							+ max_wal_senders
+#endif		/* pg12+ */
+							+ 1));
+#endif		/* pg15- */
 }
 #endif
 
@@ -967,6 +945,12 @@ pgsk_planner(Query *parse,
 
 		/* store current number of block reads and writes */
 		pgsk_entry_store(parse->queryId, PGSK_PLAN, plan_nested_level + exec_nested_level, counters);
+
+		if (pgsk_counters_hook)
+		    pgsk_counters_hook(&counters,
+							   query_string,
+							   plan_nested_level + exec_nested_level,
+							   PGSK_PLAN);
 	}
 	else
 	{
@@ -1101,6 +1085,12 @@ pgsk_ExecutorEnd (QueryDesc *queryDesc)
 
 		/* store current number of block reads and writes */
 		pgsk_entry_store(queryId, PGSK_EXEC, exec_nested_level, counters);
+
+		if (pgsk_counters_hook)
+		    pgsk_counters_hook(&counters,
+							   (const char *)queryDesc->sourceText,
+							   exec_nested_level,
+							   PGSK_EXEC);
 	}
 
 	/* give control back to PostgreSQL */
